# DbgrURI.pm -- Manage Session filenames and URIs in this module
#
# Copyright (c) 2004 - 2005 ActiveState, a division of Sophos Plc.
# All rights reserved.
# 
# This software (the Perl-DBGP package) is covered by the Artistic License
# (http://www.opensource.org/licenses/artistic-license.php).

package DB::DbgrURI;

use DB::DbgrCommon;
use DB::OSType;

use File::Spec::Functions qw(
	canonpath
	catdir
	splitpath
	splitdir
	catpath
	rel2abs
);

use strict;
use vars qw($VERSION
	    @ISA
	    @EXPORT
	    @EXPORT_OK
	    %fileURILookupTable
	    @fileURI_No_ReverseLookupTable
	    %perlNameToFileURINo
	    %mem_uriToFilename
	    %filenameToURI
	    $ldebug
	    );

$VERSION = 0.10;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
		 );
@EXPORT_OK = qw(canonicalizeFName
		    canonicalizeURI
		    filenameToURI
		    uriToFilename

		    %fileURILookupTable
		    %perlNameToFileURINo
		    );

%fileURILookupTable = ();             # Map fileURI => fileURI_No
@fileURI_No_ReverseLookupTable = ();  # Map fileURI_No => fileURI
%perlNameToFileURINo = ();            # Map perl-filename => fileURI_No

# Memoize things we look up often
%mem_uriToFilename = ();
%filenameToURI = ();

$ldebug;

# ---------------- Configuring routine

sub init {
    my %options = @_;
    $ldebug = delete $options{ldebug};
}

# ---------------- Exported Routines

sub canonicalizeFName {
    my $fname = shift;
    $fname = canonpath($fname);
    if (IS_WIN32) {
	# Map everything to one case on Windows
	$fname = lc $fname;
	$fname =~ s@\\@/@g;
    }
    return $fname;
}

# Not sure if this is still needed

sub canonicalizeURI {
    my ($bFileURI) = @_;
    my $new_URI = $bFileURI;
    $new_URI = lc $new_URI if IS_WIN32;
    my $plus_escape = '%' . sprintf('%02x', ord('+')); # lower-case
    $new_URI =~ s/\+/$plus_escape/g;
    if ($new_URI ne $bFileURI) {
	dblog("canonicalizeURI($bFileURI) ==> $new_URI");
	return $new_URI;
    }
    return $bFileURI;
}


sub filenameToURI {
    my ($bFileName) = @_;  # possibly backslashed name
    my $origFileName = $bFileName;
    return $filenameToURI{$origFileName} if $filenameToURI{$origFileName};
    # dblog("filenameToURI - 0 - [$bFileName]");
    my $start;
    $start = "filenameToURI($bFileName) => " if $ldebug;
    my $uri;
    my $canon_URI;
    my $leadingSlashes;
    if (IS_WIN32) {
	if ($bFileName =~ /^\w:[\\\/]/) {
	    # drive-letter -- colon -- path
	    $bFileName = canonpath($bFileName);
	    $bFileName = _encode_win_file_parts($bFileName);
	    $bFileName =~ s,\\,/,g;
	    $leadingSlashes = "///";
	} elsif ($bFileName =~ /^\\\\/) {
	    # It's a UNC path
	    # Remove extra "." and "x\\.." components
	    $bFileName = canonpath(substr($bFileName, 1));
	    # dblog("filenameToURI - 1 - [$bFileName // $1]");
	    $bFileName = '\\' . _encode_unix_file_parts($bFileName);
	    # dblog("filenameToURI - 2 - [$bFileName]");
	    $bFileName =~ s,\\,/,g;
	    # dblog("filenameToURI - 3 - [$bFileName]");
	    $leadingSlashes = "";
	} else {
	    $bFileName = canonpath(rel2abs($bFileName));
	    $bFileName = _encode_win_file_parts($bFileName);
	    $bFileName =~ s,\\,/,g;
	    $leadingSlashes = ($bFileName =~ /^\//) ? "//" : "///";
	}
    } else {
	# Unix pathnames go through the mount table,
	# so we never have a hostname here
	$bFileName = canonpath(rel2abs($bFileName));
	$bFileName = _encode_unix_file_parts($bFileName);
	$leadingSlashes = "//";
    }
    $canon_URI = "file:$leadingSlashes$bFileName";
    # dblog("$start$canon_URI") if $ldebug;
    $filenameToURI{$origFileName} = $canon_URI;
    return $canon_URI;
}

# This one always returns an absolute forward-slashed filename
# Even UNC paths appear as "//<host>/<path>

sub uriToFilename {
    my ($bFileURI) = @_;
    my $origURI = $bFileURI;
    return $mem_uriToFilename{$origURI} if $mem_uriToFilename{$origURI};
    # Assume all file:// URIs are local
    my $start;
    $start = "uriToFilename($bFileURI) => " if $ldebug;

    my $filename;
    if ($bFileURI =~ m@^file://(.*)$@) {
	my $file_part = _uri_decode($1);
	$filename = $file_part;
    } else {
	die "uriToFilename: Can't process URI $bFileURI";
    }
    $mem_uriToFilename{$origURI} = $filename;
    return $filename;
}

sub _encode_win_file_parts {
    my ($full_win_name) = @_;
    # dblog("_encode_win_file_parts, \$full_win_name = $full_win_name");
    my ($volume, $directories, $path) = splitpath($full_win_name);
    # dblog("_encode_win_file_parts, \$directories = $directories");
    my @dir_parts = map { _uri_encode($_) } splitdir($directories);
    # dblog("_encode_win_file_parts: split into @dir_parts");
    my $new_name = lc(catpath($volume,
			      catdir(@dir_parts),
			      _uri_encode($path)));
    # dblog("_encode_win_file_parts: map [$full_win_name] => [$new_name]");
    return $new_name;
}

# Like Windows, but this kind of filename has no initial volume

sub _encode_unix_file_parts {
    my ($full_name) = @_;
    my @dir_parts = map { _uri_encode($_) } splitdir($full_name);
    my $new_name = catdir(@dir_parts);
    # dblog("_encode_unix_file_parts: map [$full_name] => [$new_name]");
    return $new_name;
}


sub _uri_decode {
  my $todecode = shift;
  $todecode =~ tr/+/ /;       # pluses become spaces
  $todecode =~ s/%([0-9a-fA-F]{2})/chr(hex($1))/ge;
  return $todecode;
}

sub _uri_encode {
  my $toencode = shift;
  # force bytes while preserving backward compatibility -- dankogai
  $toencode = pack("C*", unpack("C*", $toencode));
  $toencode=~s/([^a-zA-Z0-9_.-])/sprintf('%%%02x', ord($1))/eg;
  return $toencode;
}

unless (caller()) {
    require Cwd;
    my $cwd = lc Cwd::getcwd();
    $cwd =~ s,\\,/,g;
    $cwd =~ s,/[^/]+$,,;
    my @fnames = (['\\\\host\\a\\b\\c\\d\\e.ext1',
		   'file://host/a/b/c/d/e.ext1'],
		  ["c:/a/z.ext2",
		   'file:///c:/a/z.ext2'],
		  ["..\\a\\b\\z.ext3",
		   "file:///$cwd/a/b/z.ext3"],
		  [
		  "/a/b",
		   "file:///a/b"
		   ],
		  [
		   "\\a\\b\\z.ext5",
		   "file:///a/b/z.ext5"
		   ],
		  );
    foreach my $obj (@fnames) {
	my $fname = $obj->[0];
	my $uri = filenameToURI($fname);
	if ($uri ne $obj->[1]) {
	    print "filenameToURI($fname) => [$uri], expecting [$obj->[1]]\n";
	}
    }
}

1;
